import hashlib


def base58encode(n: str) -> str:
    b58 = '123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz'
    result = ''
    while n > 0:
        result = b58[n % 58] + result
        n //= 58
    return result


def base256decode(s: str) -> str:
    result = 0
    for c in s:
        result = result * 256 + ord(c)
    return result


def count_leading_zeroes(s: str) -> int:
    count = 0
    for c in s:
        if c == '\0':
            count += 1
        else:
            break
    return count


def base58check_encode(prefix: int, payload: str) -> str:
    s = chr(prefix) + payload
    checksum = hashlib.sha256(s.encode()).digest()[0:4]
    result = s + checksum.hex()
    return '1' * count_leading_zeroes(result) + base58encode(base256decode(result))



import collections
import operator

from typing import Union, List

from openchain.models.block import Block
from openchain.models.exception import (
    BlockchainTreeCollisionException,
    BlockchainInvalidGenesisBlockException
)


class BlockchainNode:

    block = None
    prev_node = None
    next_nodes = []
    depth = 0

    def __init__(self, block: Block, prev_node: 'BlockchainNode'=None,
                 next_nodes: List['BlockchainNode']=None):
        self.block = block
        self.prev_node = prev_node
        if next_nodes is not None:
            self.next_nodes = next_nodes
        self.depth = 0

    @property
    def next_node(self) -> Union['BlockchainNode', None]:
        if not self.next_nodes:
            return None
        return max(self.next_nodes, key=operator.attrgetter('depth'))

    @property
    def __dict__(self) -> dict:
        prev_block = None
        if self.prev_node and self.prev_node.block:
            prev_block = self.prev_node.block.data_hash

        next_block = None
        if self.next_node and self.next_node.block:
            next_block = self.next_node.block.data_hash

        unordered = {
            'block': self.block.__dict__,
            'prev_block': prev_block,
            'next_block': next_block,
        }
        return collections.OrderedDict(sorted(unordered.items()))

    def calculate_depth(self) -> int:
        if self.next_nodes:
            self.depth = max(
                map(lambda x: x.calculate_depth(), self.next_nodes)
            ) + 1
        return self.depth


class Blockchain:

    block_tree = {}
    block_list = []
    is_tree_generated = True

    def __init__(self, block_list: List[Block]):
        self.block_tree = {}
        self.block_list = block_list
        self.is_tree_generated = False

    @property
    def __dict__(self) -> dict:
        result = {}
        for block_hash, block in self.block_tree.items():
            result[block_hash] = block.__dict__
        return result

    @property
    def last_block_hash(self) -> Union[str, None]:
        block_keys = list(self.block_tree.keys())
        return block_keys[-1] if len(block_keys) else None

    def generate_nodes(self, current_block: Block, prev_block: Block=None) -> BlockchainNode:
        next_nodes = []
        for block in self.block_list:
            if block.prev_block == current_block.data_hash:
                next_node = self.generate_nodes(block, current_block)
                if next_node is not None:
                    next_nodes.append(next_node)

        prev_node = BlockchainNode(prev_block)
        return BlockchainNode(current_block, prev_node, next_nodes)

    def generate_tree(self, current_block: BlockchainNode, raise_exception: bool=False):
        if current_block is None:
            return

        self.block_tree[current_block.block.data_hash] = current_block

        if len(current_block.next_nodes) > 1 and raise_exception:
            raise BlockchainTreeCollisionException

        self.generate_tree(current_block.next_node, raise_exception)

    def build(self, raise_exception: bool=True, genesis_block_hash: str=None):
        if self.is_tree_generated:
            return

        # TODO: Add tests
        genesis_block = None
        for genesis_block in self.block_list:
            if genesis_block_hash and genesis_block.data_hash == genesis_block_hash:
                break
            if genesis_block.is_genesis:
                break

        if genesis_block is None:
            if raise_exception:
                raise BlockchainInvalidGenesisBlockException
            else:
                return

        genesis_node = self.generate_nodes(genesis_block)
        self.generate_tree(genesis_node, raise_exception)

        self.is_tree_generated = True

        for key, node in self.block_tree.items():
            node.depth = node.calculate_depth()

    @property
    def is_valid(self):
        try:
            self.build(True)
            return True
        except BlockchainTreeCollisionException:
            return False

    def can_add_block(self, block: object) -> bool:
        # Store original blockchain tree
        original_block_tree = self.block_tree

        # Store original blockchain tree
        self.block_list.append(block)
        is_valid = self.is_valid

        # Restore original blockchain tree
        self.block_tree = original_block_tree
        return is_valid



import os


class Singleton(type):

    instance_set = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls.instance_set:
            cls.instance_set[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls.instance_set[cls]


class BaseAdapter(metaclass=Singleton):

    db_path = os.getenv('DATABASE_PATH', '/var/lib/openchain/')
    connection_set = {}

    def connect(self, namespace: str):
        raise NotImplementedError

    def get(self, namespace: str, key: bytes) -> bytes:
        raise NotImplementedError

    def put(self, namespace, key: bytes, value: bytes):
        raise NotImplementedError

    def delete(self, namespace: str, key: bytes) -> bytes:
        raise NotImplementedError

    def iterator(self, namespace: str) -> callable:
        raise NotImplementedError

    def batch_put(self, namespace: str, item_list: list) -> callable:
        raise NotImplementedError

    def batch_delete(self, namespace: str, item_list: list) -> callable:
        raise NotImplementedError



class ModelFactory:

    @staticmethod
    def get_model(name):
        from openchain.models.block import Block
        from openchain.models.blockchain import Blockchain
        from openchain.models.client import Client
        from openchain.models.transaction import Transaction
        from openchain.models.wallet import Wallet

        model_map = {
            'block': Block,
            'blockchain': Blockchain,
            'client': Client,
            'transaction': Transaction,
            'wallet': Wallet,
        }

        if name in model_map:
            return model_map[name]
        return None


class BlockchainFactory:

    @staticmethod
    def build_blockchain(block_list, genesis_block_hash: str=None):
        blockchain = ModelFactory.get_model('blockchain')(block_list)
        blockchain.build(raise_exception=False,
                         genesis_block_hash=genesis_block_hash)
        return blockchain



import collections
import time

from openchain.models.base import Model, Manager


class ClientManager(Manager):

    def append(self, item: object, commit: bool=False):
        updated = False
        for c in self.get():
            if item.client_id == c.client_id:
                updated = True
                item.timestamp = time.time()
        if not updated:
            self.queryset.append(item)
        self.save()


class Client(Model):

    client_id = None
    timestamp = None

    objects = ClientManager()

    def __init__(self, client_id: str, timestamp: float=None):
        self.client_id = client_id
        if timestamp is None:
            self.timestamp = time.time()

    @property
    def __dict__(self):
        unordered = {
            'client_id': self.client_id,
            'timestamp': self.timestamp
        }
        return collections.OrderedDict(sorted(unordered.items()))



import json
import time
import collections

from hashlib import sha256

from openchain.models.base import Model, Manager
from openchain.models.exception import BlockInvalidException
from openchain.models.factory import ModelFactory, BlockchainFactory


class BlockManager(Manager):

    def search(self, search_hash):
        if not self.loaded:
            self.load()
        for item in self.queryset:
            if search_hash == item.data_hash:
                return item

    def append(self, block: callable, commit: bool=False):
        prev_block = self.search(block.prev_block)
        if (prev_block is not None or block.is_genesis) and block.is_valid:
            if not block.is_genesis:
                prev_block.next_block = block.data_hash
                prev_block.save()
            self.queryset.append(block)
            self.save()


class Block(Model):

    prev_block = None
    next_block = None
    data_hash = None
    nonce = 0
    transactions = []
    timestamp = None

    objects = BlockManager()

    def __init__(self, prev_block: str=None, next_block: str=None, data_hash: str=None,
                 nonce: int=0, transactions: list=None, timestamp: float=None):
        self.prev_block = prev_block
        if next_block is not None:
            self.next_block = next_block
        if data_hash is not None:
            self.data_hash = data_hash
        if nonce is not None:
            self.nonce = nonce
        if transactions is not None:
            self.transactions = []
            for transaction in transactions:
                self.transactions.append(ModelFactory.get_model('transaction')(**transaction))
        if timestamp is not None:
            self.timestamp = timestamp

    @property
    def is_genesis(self) -> bool:
        """
        Return True if current block is genesis block
        :returns True if correct, False if not
        """
        return self.prev_block is None

    @property
    def data(self) -> bytes:
        """
        Return bytes representation of hashable data for Block object
        :returns bytes
        """
        data = {
            'prev': self.prev_block,
            'transactions': [tx.__dict__ for tx in self.transactions],
            'timestamp': self.timestamp
        }
        ordered = collections.OrderedDict(sorted(data.items()))
        return json.dumps(ordered).encode()

    @property
    def __dict__(self):
        """
        Return dict representation for Block model
        :raises BlockInvalidException: if block has invalid hash
        :returns OrderedDict
        """
        if not self.is_valid:
            raise BlockInvalidException
        unordered = {
            'prev_block': self.prev_block,
            'next_block': self.next_block,
            'data_hash': self.data_hash,
            'nonce': self.nonce,
            'transactions': [tx.__dict__ for tx in self.transactions],
            'timestamp': self.timestamp
        }
        return collections.OrderedDict(sorted(unordered.items()))

    def generate(self):
        """
        Simple Proof of Work Algorithm:
         - Find a number p' such that hash(pp') contains leading 4 zeroes, where p is the data hash
         - p is the data hash, and p' is the new proof
        """

        if self.timestamp is None:
            self.timestamp = time.time()

        self.data_hash = sha256(sha256(self.data).digest()).digest().hex()
        while self.valid_nonce(self.nonce) is False:
            self.nonce += 1

    def valid_nonce(self, nonce: int) -> bool:
        """
        Validates the Nonce
        :param nonce: Current Nonce
        :returns: True if correct, False if not.
        """

        guess = '{}{}'.format(self.data_hash, nonce).encode()
        guess_hash = sha256(guess).hexdigest()
        return guess_hash[:4] == '0000'

    @property
    def is_valid(self) -> bool:
        """
        Validates block hash
        :returns True if correct, False if not
        """

        hashed_raw_block = sha256(sha256(self.data).digest()).digest().hex()
        if self.data_hash != hashed_raw_block:
            return False
        return self.valid_nonce(self.nonce)

    def save(self):
        """
        Validate and save current block
        :raises BlockInvalidException: if block has invalid hash
        :raises BlockchainTreeChildCollisionException: if block has collision with child block
        :raises BlockchainTreeParentCollisionException: if block has collision with parent block
        """

        if not self.is_valid:
            raise BlockInvalidException

        # Check ability to add block to blockchain
        blockchain = BlockchainFactory.build_blockchain(self.objects.get())
        if blockchain.can_add_block(self):
            # Update parent block
            self.objects.append(self, True)

            # Save current block
            self.objects.append(self, True)




class BlockchainTreeCollisionException(Exception):
    """Blockchain node has two different childs"""


class BlockchainInvalidGenesisBlockException(Exception):
    """Blockchain node has two different parents"""


class BlockInvalidException(Exception):
    """Block is invalid"""


class TransactionInvalidPublicKeyException(Exception):
    """Transaction has invalid public key"""


class TransactionInvalidSignatureException(Exception):
    """Transaction has invalid signature"""



import os
import leveldb

from openchain.adapters.base import BaseAdapter


class LevelDBAdapter(BaseAdapter):

    def connect(self, namespace):
        if namespace not in self.connection_set:
            filename = os.path.join(self.db_path, namespace)
            self.connection_set[namespace] = leveldb.LevelDB(filename, create_if_missing=False)

    def get(self, namespace, key):
        return self.connection_set[namespace].Get(key)

    def put(self, namespace, key, value):
        self.connection_set[namespace].Put(key, value)

    def delete(self, namespace, key):
        self.connection_set[namespace].Delete(key)

    def iterator(self, namespace):
        return self.connection_set[namespace].RangeIter()

    def batch_put(self, namespace, items):
        with self.connection_set[namespace].WriteBatch() as batch:
            for key, value in [i.items() for i in items]:
                batch.Put(key, value)
            self.connection_set[namespace].Write(batch)



#!/usr/bin/env python
import io
import os

from setuptools import setup, find_packages

from openchain import __version__


def read(file_name):
    return io.open(os.path.join(os.path.dirname(__file__), file_name), encoding='utf-8').read()


exec(read('openchain/__init__.py'))

setup(
    name='openchain',
    version=__version__,
    description='Openchain Library',
    long_description=read('README.rst'),
    url='https://bitbucket.org/manti_by/openchain/',
    author='Alexander Chaika',
    author_email='manti.by@gmail.com',
    license='BSD',
    python_requires='>=3.6',
    packages=find_packages(exclude=['*.tests']),
    install_requires=[
        'ecdsa>=0.13,<0.14'
        'leveldb>=0.194,<0.195',
        'plyvel>=1.0.5,<1.1.0',
        'xxhash>=1.2.0,<1.3.0'
    ],
    test_suite='tests',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: BSD License',
        'Programming Language :: Python :: 3.6',
        'Operating System :: POSIX :: Linux',
    ],
    keywords='blockchain',
)



import os
import plyvel

from openchain.adapters.base import BaseAdapter


class PlyvelAdapter(BaseAdapter):

    def connect(self, namespace):
        if namespace not in self.connection_set:
            filename = os.path.join(self.db_path, namespace)
            self.connection_set[namespace] = plyvel.DB(filename, create_if_missing=True)

    def get(self, namespace, key):
        return self.connection_set[namespace].get(key)

    def put(self, namespace, key, value):
        self.connection_set[namespace].put(key, value)

    def delete(self, namespace, key):
        self.connection_set[namespace].delete(key)

    def iterator(self, namespace):
        return self.connection_set[namespace].iterator()

    def batch_put(self, namespace, item_list):
        with self.connection_set[namespace].write_batch() as batch:
            for item in item_list:
                for key, value in item.items():
                    batch.put(key, value)
            batch.write()

    def batch_delete(self, namespace, item_list):
        with self.connection_set[namespace].write_batch() as batch:
            for item in item_list:
                batch.delete(next(iter(item)))
            batch.write()



import collections
import hashlib

from ecdsa import VerifyingKey, SigningKey, SECP256k1

from openchain.utils.base import base58check_encode
from openchain.models.base import Model, Manager


class WalletManager(Manager):

    pass


class Wallet(Model):

    private_key = None
    public_key = None

    objects = WalletManager()

    def __init__(self, private_key: str=None, public_key: str=None):
        if private_key is None:
            self.private_key = SigningKey.generate(curve=SECP256k1)
        else:
            self.private_key = SigningKey.from_string(bytes.fromhex(private_key), curve=SECP256k1)

        if public_key is None:
            self.public_key = self.private_key.get_verifying_key()
        else:
            self.public_key = VerifyingKey.from_string(bytes.fromhex(public_key), curve=SECP256k1)

    @property
    def __dict__(self) -> dict:
        unordered = {
            'private_key': self.private_key.to_string().hex(),
            'public_key': self.public_key.to_string().hex()
        }
        return collections.OrderedDict(sorted(unordered.items()))

    @property
    def private_key_hex(self) -> str:
        return self.private_key.to_string().hex()

    @property
    def private_key_bytes(self) -> bytes:
        return bytes.fromhex(self.private_key_hex)

    @property
    def public_key_hex(self) -> str:
        return self.public_key.to_string().hex()

    @property
    def public_key_bytes(self) -> bytes:
        return bytes.fromhex(self.public_key_hex)

    @property
    def address(self) -> str:
        ripemd160 = hashlib.new('ripemd160')
        ripemd160.update(hashlib.sha256(self.public_key_bytes).digest())
        return base58check_encode(0, ripemd160.digest().hex())



import json
import xxhash

from openchain.adapters.plyvel import PlyvelAdapter
from openchain.models.factory import ModelFactory


class Manager:

    db_adapter = PlyvelAdapter()

    def __init__(self):
        self.loaded = False
        self.queryset = []
        self.namespace = self.__class__.__name__.lower()
        if self.namespace != 'manager':
            self.namespace = self.namespace.replace('manager', '')
        self.db_adapter.connect(self.namespace)

    def load(self, force_load=False):
        if not self.loaded or force_load:
            self.queryset = []
            for key, value in self.db_adapter.iterator(self.namespace):
                data = json.loads(value.decode())
                self.queryset.append(ModelFactory.get_model(self.namespace)(**data))
            self.loaded = True

    def get(self) -> list:
        self.load()
        return self.queryset

    def set(self, qs: list):
        self.delete_all()
        self.queryset = qs
        self.save()

    def search(self, search_item):
        self.load()
        for item in self.queryset:
            if search_item == item:
                return item

    def append(self, item: object, commit: bool=False):
        self.load()
        self.queryset.append(item)
        if commit:
            self.save()

    def delete(self, item: object):
        if item in self.queryset:
            self.queryset.remove(item)
            index = next(iter(item.__dict__)).encode()
            self.db_adapter.delete(self.namespace, index)

    def delete_list(self, item_list: list):
        hashed_list = []
        for item in item_list:
            index = xxhash.xxh64()
            index.update(item.__bytes__)
            hashed_list.append({index.digest(): item.__bytes__})

        self.db_adapter.batch_delete(self.namespace, hashed_list)
        self.load(True)

    def delete_all(self):
        self.load()
        self.db_adapter.batch_delete(self.namespace, self.hashed_list)
        self.queryset = []

    def save(self):
        self.db_adapter.batch_put(self.namespace, self.hashed_list)

    @property
    def hashed_list(self):
        result = []
        if self.queryset:
            for item in self.queryset:
                index = xxhash.xxh64()
                index.update(item.__bytes__)
                result.append({index.digest(): item.__bytes__})
        return result

    @property
    def dict_list(self):
        result = []
        for item in self.queryset:
            result.append(item.__dict__)
        return result


class Model:

    objects = Manager()

    @property
    def __dict__(self) -> dict:
        raise NotImplementedError()

    @property
    def __str__(self) -> str:
        return json.dumps(self.__dict__)

    @property
    def __bytes__(self) -> bytes:
        return self.__str__.encode()

    def save(self):
        self.objects.append(self, True)



import json
import collections

from hashlib import sha256
from ecdsa import SigningKey, VerifyingKey, SECP256k1
from ecdsa.keys import BadSignatureError

from openchain.models.base import Model, Manager
from openchain.models.exception import TransactionInvalidPublicKeyException


class TransactionManager(Manager):

    pass


class Transaction(Model):

    in_address = None
    out_address = None
    amount = None
    data_hash = None
    public_key = None
    signature = None

    objects = TransactionManager()

    def __init__(self, in_address: str, out_address: str, amount: float,
                 public_key: str=None, signature: str=None):
        self.in_address = in_address
        self.out_address = out_address
        self.amount = amount
        self.data_hash = sha256(sha256(self.data).digest()).digest()

        if public_key is not None:
            self.public_key = VerifyingKey.from_string(bytes.fromhex(public_key), curve=SECP256k1)
        if signature is not None:
            self.signature = bytes.fromhex(signature)

    @property
    def data(self) -> bytes:
        data = {
            'in_address': self.in_address,
            'out_address': self.out_address,
            'amount': self.amount
        }
        ordered = collections.OrderedDict(sorted(data.items()))
        return json.dumps(ordered).encode()

    @property
    def __dict__(self) -> dict:
        unordered = {
            'in_address': self.in_address,
            'out_address': self.out_address,
            'amount': self.amount,
            'public_key': self.public_key.to_string().hex() if self.public_key else None,
            'signature': self.signature.hex()
        }
        return collections.OrderedDict(sorted(unordered.items()))

    def signing(self, private_key: str):
        self.data_hash = sha256(sha256(self.data).digest()).digest()
        signing_key = SigningKey.from_string(bytes.fromhex(private_key), curve=SECP256k1)
        if self.public_key is None:
            self.public_key = signing_key.verifying_key
        elif self.public_key.to_string() != signing_key.verifying_key.to_string():  # check pre-assigned key
            raise TransactionInvalidPublicKeyException
        if self.signature is None:
            self.signature = signing_key.sign(self.data_hash)

    @property
    def is_valid(self) -> bool:
        if self.public_key is not None and self.signature is not None:
            try:
                hashed_raw_transaction = sha256(sha256(self.data).digest()).digest()
                if self.data_hash != hashed_raw_transaction:
                    return False
                return self.public_key.verify(self.signature, hashed_raw_transaction)
            except BadSignatureError:
                pass
        return False
